import typing
from dramatiq import get_broker, Message
from pypeline.pipelines.composition.parallel_pipeline_composition import (
    parallel_pipeline,
)
from pypeline.dramatiq import LazyActor
from pypeline.utils.dramatiq_utils import register_lazy_actor
from pypeline.pipeline_settings_schema import (
    MissingSettingsException,
    create_pipeline_settings_schema,
    PipelineScenarioSchema,
)
from pypeline.pipelines.composition.pypeline_composition import Pypeline
from pypeline.utils.config_utils import retrieve_latest_pipeline_config
from pypeline.utils.module_utils import get_callable
from pypeline.utils.pipeline_utils import (
    get_execution_graph,
    topological_sort_with_parallelism,
)


def dag_generator(
    pipeline_id: str, scenarios: typing.List[typing.Dict] = [], *args, **kwargs
) -> typing.Union[parallel_pipeline, Pypeline]:
    """Generates a pipeline dag from a pre-defined pipeline yaml

    :param pipeline_id: Id of the pipeline to generate
    :param task_replacements: A dictionary of task names and handler index to run. E.g. {"a": 1} would run the handler
        in the second index position.
    :param scenarios:
    :param args:
    :param kwargs:
    :return: Returns a parallel_pipeline object which can be run
    """
    pipeline = retrieve_latest_pipeline_config(pipeline_id=pipeline_id)

    pipeline_config = pipeline["config"]
    broker = get_broker()
    broker.actors.clear()

    if pipeline["schemaVersion"] == 2:
        supplied_pipeline_settings_schema = create_pipeline_settings_schema(
            pipeline_config["settings"]
        )

        # Validate scenarios settings to make sure they look okay
        validated_scenarios = PipelineScenarioSchema(many=True).load(scenarios)

        for scenario in validated_scenarios:
            supplied_pipeline_settings_schema.load(scenario["settings"])

        p = Pypeline(pipeline, scenarios=scenarios, broker=broker)
        return p

    graph = get_execution_graph(pipeline_config)
    optimal_execution_graph = topological_sort_with_parallelism(graph.copy())
    registered_actors: typing.Dict[str, LazyActor] = {}

    messages: typing.List[typing.List[Message]] = []

    task_definitions = pipeline_config["taskDefinitions"]
    for task_group in optimal_execution_graph:
        message_group = []
        for task in task_group:
            module_path = task_definitions[task]["handler"]
            server_type = task_definitions[task].get("serverType", None)
            tmp_handler = get_callable(module_path)
            lazy_actor = register_lazy_actor(
                broker, tmp_handler, pipeline_config["metadata"], server_type
            )
            registered_actors[task] = lazy_actor
            if args and not kwargs:
                msg = registered_actors[task].message(*args)
            elif kwargs and not args:
                msg = registered_actors[task].message(**kwargs)
            elif args and kwargs:
                msg = registered_actors[task].message(*args, **kwargs)
            else:
                msg = registered_actors[task].message()
            if pipeline_config["metadata"].get("maxRetry", None) is not None:
                msg.options["max_retries"] = pipeline_config["metadata"]["maxRetry"]
            msg.options["task_ttl"] = pipeline_config["metadata"]["maxTtl"]
            message_group.append(msg)

        messages.append(message_group)
    p = parallel_pipeline(messages)

    return p



""" Flask specific decorators (primarily for auth activities and app context)
"""
import os
import logging
from http import HTTPStatus
from functools import wraps
from flask import request
from flask import abort

logger = logging.getLogger(__name__)


def require_accesskey(fn):
    """Convenience decorator to add to a web route (typically an API)
    when using Flask.

    Usage::
        from sermos import Blueprint, ApiServices
        bp = Blueprint('api_routes', __name__, url_prefix='/api')

        @bp.route('/my-api-route')
        class ApiClass(MethodView):
            @require_access_key
            def post(self, payload: dict):
                return {}
    """

    @wraps(fn)
    def decorated_view(*args, **kwargs):
        access_key = request.headers.get("accesskey")
        if not access_key:
            access_key = request.args.get("accesskey")

        configured_access_key = os.environ.get("API_ACCESS_KEY", None)

        if access_key == configured_access_key:
            return fn(*args, **kwargs)

        abort(HTTPStatus.UNAUTHORIZED)

    return decorated_view



import copy
import typing
import pika
import logging
import os

import click
from urllib.parse import urlparse

from dramatiq.brokers.redis import RedisBroker
from redis.sentinel import Sentinel
from pypeline.extensions import pypeline_config
from warnings import warn
from apscheduler.schedulers.blocking import BlockingScheduler
from apscheduler.triggers.cron import CronTrigger
from dramatiq import Broker, Middleware, set_broker, get_broker
from dramatiq.brokers.rabbitmq import RabbitmqBroker
from dramatiq.cli import (
    CPUS,
    HAS_WATCHDOG,
    main as dramatiq_worker,
    make_argument_parser as dramatiq_argument_parser,
    import_object,
)
from dramatiq.middleware import default_middleware, CurrentMessage
from dramatiq.results import Results
from dramatiq.results.backends.redis import RedisBackend
from flask import current_app, Flask
from flask.cli import with_appcontext

from pypeline.constants import (
    REDIS_URL,
    REDIS_SENTINEL_MASTER_NAME,
    RABBIT_URL,
    DEFAULT_BROKER_CALLABLE,
    DEFAULT_BROKER_CONNECTION_HEARTBEAT,
    DEFAULT_BROKER_BLOCKED_CONNECTION_TIMEOUT,
    DEFAULT_BROKER_CONNECTION_ATTEMPTS,
    MESSAGE_BROKER,
    DEFAULT_BROKER_HEARTBEAT_TIMEOUT,
    DEFAULT_REDIS_SOCKET_CONNECT_TIMEOUT,
    DEFAULT_REDIS_SOCKET_TIMEOUT,
    DEFAULT_REDIS_RETRY_ON_TIMEOUT,
    DEFAULT_REDIS_SOCKET_KEEPALIVE,
    DEFAULT_REDIS_HEALTH_CHECK_INTERVAL,
)
from pypeline.pipelines.middleware.get_active_worker_id_middleware import (
    GetActiveWorkerIdMiddleware,
)
from pypeline.pipelines.middleware.parallel_pipeline_middleware import ParallelPipeline
from pypeline.pipelines.middleware.pypeline_middleware import PypelineMiddleware
from pypeline.utils.config_utils import (
    retrieve_latest_schedule_config,
    get_service_config_for_worker,
    retrieve_executable_job_config,
)
from pypeline.utils.dramatiq_utils import (
    guess_code_directory,
    list_managed_actors,
    register_lazy_actor,
    LazyActor,
)
from pypeline.utils.graceful_shutdown_util import enable_graceful_shutdown
from pypeline.utils.module_utils import get_callable
from dramatiq.middleware import (
    Retries,
    Callbacks,
    TimeLimit,
    AgeLimit,
    ShutdownNotifications,
    Pipelines,
)

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def configure_default_broker(broker: Broker = None):
    reworked_defaults=[AgeLimit(), TimeLimit(), ShutdownNotifications(), Callbacks(), Pipelines(), Retries()]
    redis_client = None
    if REDIS_SENTINEL_MASTER_NAME is not None:
        parsed_redis_url = urlparse(REDIS_URL)
        redis_sentinel = Sentinel(
            sentinels=[(parsed_redis_url.hostname, parsed_redis_url.port)],
        )
        redis_client = redis_sentinel.master_for(
            REDIS_SENTINEL_MASTER_NAME,
            db=int(parsed_redis_url.path[1]) if parsed_redis_url.path else 0,
            password=parsed_redis_url.password,
            socket_connect_timeout=DEFAULT_REDIS_SOCKET_CONNECT_TIMEOUT,
            socket_timeout=DEFAULT_REDIS_SOCKET_TIMEOUT,
            retry_on_timeout=DEFAULT_REDIS_RETRY_ON_TIMEOUT,
            socket_keepalive=DEFAULT_REDIS_SOCKET_KEEPALIVE,
            health_check_interval=DEFAULT_REDIS_HEALTH_CHECK_INTERVAL,
        )
    redis_backend = RedisBackend(client=redis_client, url=REDIS_URL)

    if MESSAGE_BROKER == "RABBITMQ":
        parsed_url = urlparse(RABBIT_URL)
        credentials = pika.PlainCredentials(parsed_url.username, parsed_url.password)
        broker = (
            broker
            if broker is not None
            else RabbitmqBroker(
                host=parsed_url.hostname,
                port=parsed_url.port,
                credentials=credentials,
                heartbeat=DEFAULT_BROKER_CONNECTION_HEARTBEAT,
                connection_attempts=DEFAULT_BROKER_CONNECTION_ATTEMPTS,
                blocked_connection_timeout=DEFAULT_BROKER_BLOCKED_CONNECTION_TIMEOUT,
                middleware=reworked_defaults
            )
        )

    elif MESSAGE_BROKER == "REDIS":
        broker = (
            broker
            if broker is not None
            else RedisBroker(
                client=redis_client,
                url=REDIS_URL,
                heartbeat_timeout=DEFAULT_BROKER_HEARTBEAT_TIMEOUT,
                middleware=reworked_defaults
            )
        )

    broker.add_middleware(Results(backend=redis_backend))
    broker.add_middleware(ParallelPipeline(redis_url=REDIS_URL))
    broker.add_middleware(PypelineMiddleware(redis_url=REDIS_URL))
    broker.add_middleware(GetActiveWorkerIdMiddleware())
    broker.add_middleware(CurrentMessage())
    if (
        os.getenv("RESTRICT_WORKER_SHUTDOWN_WHILE_JOBS_RUNNING", "false").lower()
        == "true"
    ):
        enable_graceful_shutdown(broker=broker, redis_url=REDIS_URL)

    register_actors_for_workers(broker)

    set_broker(broker)


def register_actors_for_workers(broker: Broker):
    service = get_service_config_for_worker(pypeline_config)
    scheduled_jobs_config = retrieve_latest_schedule_config()
    executable_jobs_config = retrieve_executable_job_config()

    if not service:
        return

    worker_registered_tasks = [
        task_handler["handler"] for task_handler in service.get("registeredTasks")
    ]

    # Loop over the pipelines to get metadata and other information about the task for registration
    for pipeline_key, pipeline in pypeline_config["pipelines"].items():
        for task, task_handler_meta in pipeline["config"]["taskDefinitions"].items():
            if pipeline["schemaVersion"] == 1:
                # Check if any task in this pipeline is registered
                task_handlers = [task_handler_meta["handler"]]
            elif pipeline["schemaVersion"] == 2:
                task_handlers = [t for t in task_handler_meta["handlers"]]

            for task_handler in task_handlers:
                if task_handler in worker_registered_tasks:
                    server_type = task_handler_meta.get("serverType", None)

                    try:
                        pipeline_metadata = copy.deepcopy(
                            pipeline["config"]["metadata"]
                        )
                        tmp_handler = get_callable(task_handler)
                        if pipeline_metadata.get("maxRetry", 0) >= 0:
                            pipeline_metadata["store_results"] = True
                            _ = register_lazy_actor(
                                broker, tmp_handler, pipeline_metadata, server_type
                            )
                    except Exception as e:
                        logger.exception(
                            f"Unable to add a task {task_handler} to dramatiq: {e}"
                        )
    # Loop over the scheduled jobs and create metadata and other information about the task for registration
    for job in scheduled_jobs_config:
        config = job["config"]
        if config["task"] in worker_registered_tasks:
            pipeline_meta = {"queue": config.get("queue", "default")}
            try:
                tmp_handler = get_callable(config["task"])
                if pipeline_meta and pipeline_meta.get("maxRetry", 0) >= 0:
                    pipeline_meta["store_results"] = True
                    _ = register_lazy_actor(broker, tmp_handler, pipeline_meta, None)
            except Exception as e:
                logger.exception(f"Unable to add a task to dramatiq: {e}")

    for job in executable_jobs_config or []:
        config = job["config"]
        if config["task"] in worker_registered_tasks:
            pipeline_meta = {"queue": config.get("queue", "default")}
            try:
                tmp_handler = get_callable(config["task"])
                if pipeline_meta and pipeline_meta.get("maxRetry", 0) >= 0:
                    pipeline_meta["store_results"] = True
                    _ = register_lazy_actor(broker, tmp_handler, pipeline_meta, None)
            except Exception as e:
                logger.exception(f"Unable to add a task to dramatiq: {e}")


class Dramatiq:
    """Flask extension bridging Dramatiq broker and Flask app.

    Dramatiq API is eager. Broker initialisation precede actor declaration.
    This breaks application factory pattern and other way to initialize
    configuration after import.

    This class enables lazy initialization of Dramatiq. Actual Dramatiq broker
    is instanciated only once Flask app is created.

    .. automethod:: actor
    .. automethod:: init_app
    """

    def __init__(
        self,
        app: Flask = None,
        name: str = "dramatiq",
        config_prefix: str = None,
        middleware: typing.List[Middleware] = None,
    ):
        """
        :app: Flask application if created. See :meth:`init_app`.

        :param broker_configuration_callable_module: In order to work in fork and spawn mode
            we need to configure our broker using a callable function.  Default is specified as
            "pypeline.flask_dramatiq:configure_default_broker".  This allows the user to
            override if necessary.

        :param name: Unique identifier for multi-broker app.

        :param config_prefix: Flask configuration option prefix for this
            broker. By default, it is derived from ``name`` parameter,
            capitalized.

        :param middleware: List of Dramatiq middleware instances to override
             Dramatiq defaults.

        Flask-Dramatiq always prepend a custom middleware to the middleware
        stack that setup Flask context. This way, every middleware can use
        Flask app context.

        """
        self.actors = []
        self.app = None
        self.config_prefix = config_prefix or name.upper() + "_BROKER"
        self.name = name
        self.broker = None
        if middleware is None:
            middleware = [m() for m in default_middleware]
        self.middleware = middleware
        if app:
            self.init_app(app)

    def __repr__(self) -> str:
        return "<%s %s>" % (self.__class__.__name__, self.name)

    def init_app(self, app: Flask):
        """Initialize extension for one Flask application

        This method triggers Dramatiq broker instantiation and effective actor
        registration.

        """
        if self.app is not None:
            warn(
                "%s is used by more than one flask application. "
                "Actor's context may be set incorrectly." % (self,),
                stacklevel=2,
            )
        self.app = app
        app.extensions["dramatiq-" + self.name] = self

        module_name, broker_or_callable = import_object(DEFAULT_BROKER_CALLABLE)

        # Callable function is expected to setBroker()
        if callable(broker_or_callable):
            logger.info(f"Configuring broker via {DEFAULT_BROKER_CALLABLE}")
            broker_or_callable()
        else:
            raise TypeError("DEFAULT_BROKER_CALLABLE must point to a callable function")
        self.broker = get_broker()
        for actor in self.actors:
            actor.register(broker=self.broker)

    def actor(self, fn=None, **kw):
        """Register a callable as Dramatiq actor.

        This decorator lazily register a callable as a Dramatiq actor. The
        actor can't be called before :meth:`init_app` is called.

        :param kw: Keywords argument passed to :func:`dramatiq.actor`.

        """
        # Substitute dramatiq.actor decorator to return a lazy wrapper. This
        # allows to register actors in extension before the broker is
        # effectively configured by init_app.

        def decorator(fn):
            lazy_actor = LazyActor(self, fn, kw)
            self.actors.append(lazy_actor)
            if self.app:
                lazy_actor.register(self.broker)
            return lazy_actor

        if fn:
            return decorator(fn)
        return decorator


@click.command("cron-scheduler")
def cron_scheduler():  # pragma: no cover
    # Configure our broker that we will schedule registered tasks for
    scheduler = BlockingScheduler()
    module_name, broker_or_callable = import_object(DEFAULT_BROKER_CALLABLE)

    # Callable function is expected to setBroker()
    if callable(broker_or_callable):
        logger.info(f"Configuring broker via {DEFAULT_BROKER_CALLABLE}")
        broker_or_callable()
    else:
        raise TypeError("DEFAULT_BROKER_CALLABLE must point to a callable function")

    broker = get_broker()
    jobs = retrieve_latest_schedule_config()

    for job in jobs:
        if job["enabled"]:
            config = job["config"]
            worker_path = config["task"]
            tmp_handler = get_callable(worker_path)
            pipeline_meta = {"queue": config.get("queue", "default")}
            actor = register_lazy_actor(broker, tmp_handler, pipeline_meta)
            schedule = config["schedule"]
            scheduler.add_job(
                actor.send,
                CronTrigger.from_crontab(
                    f"{schedule['minute']} {schedule['hour']} {schedule['dayOfMonth']} {schedule['monthOfYear']} {schedule['dayOfWeek']}"
                ),
            )

    try:
        scheduler.start()
    except KeyboardInterrupt:
        scheduler.shutdown()


@click.command("pypeline-worker")
@click.argument("broker_name", default="dramatiq")
@click.option(
    "-v", "--verbose", default=0, count=True, help="turn on verbose log output"
)
@click.option(
    "-p",
    "--processes",
    default=CPUS,
    metavar="PROCESSES",
    show_default=True,
    help="the number of worker processes to run",
)
@click.option(
    "-t",
    "--threads",
    default=8,
    metavar="THREADS",
    show_default=True,
    help="the number of worker treads per processes",
)
@click.option(
    "-Q",
    "--queues",
    type=str,
    default=None,
    metavar="QUEUES",
    show_default=True,
    help="listen to a subset of queues, comma separated",
)
@click.option(
    "--use-spawn",
    type=bool,
    default=False,
    metavar="USE_SPAWN",
    show_default=True,
    help="start processes by spawning (default: fork on unix, spawn on windows)",
)
@with_appcontext
def pypeline_worker(
    verbose, processes, threads, queues, broker_name, use_spawn
):  # pragma: no cover
    """Run dramatiq workers.

    Setup Dramatiq with broker and task modules from Flask app.

    \b
    examples:
      # Run dramatiq with 1 thread per process.
      $ flask worker --threads 1

    \b
      # Listen only to the "foo" and "bar" queues.
      $ flask worker -Q foo,bar

    \b
      # Consuming from a specific broker
      $ flask worker mybroker
    """
    # Plugin for flask.commands entrypoint.
    #
    # Wraps dramatiq worker CLI in a Flask command. This is private API of
    # dramatiq.

    def format_actor(actor):
        return "%s@%s" % (actor.actor_name, actor.queue_name)

    parser = dramatiq_argument_parser()

    # Set worker broker globally.
    needle = "dramatiq-" + broker_name
    broker = current_app.extensions[needle].broker
    set_broker(broker)

    command = [
        "--processes",
        str(processes),
        "--threads",
        str(threads),
        # Fall back to flask_dramatiq global broker
        DEFAULT_BROKER_CALLABLE,
    ]

    if use_spawn:
        command += ["--use-spawn"]

    if current_app.config["DEBUG"]:
        verbose = max(1, verbose)
        if HAS_WATCHDOG:
            command += ["--watch", guess_code_directory(broker)]

    queues = queues.split(",") if queues else []
    if queues:
        command += ["--queues"] + queues
    command += verbose * ["-v"]
    args = parser.parse_args(command)
    logger.info("Able to execute the following actors:")
    for actor in list_managed_actors(broker, queues):
        logger.info("    %s.", format_actor(actor))

    dramatiq_worker(args)



""" Utilities for loading modules/callables based on strings.
"""

import re
import logging
import importlib
from typing import Callable

logger = logging.getLogger(__name__)


class PypelineModuleLoader(object):
    """Helper class to load modules / classes / methods based on a path string."""

    def get_module(self, resource_dot_path: str):
        """Retrieve the module based on a 'resource dot path'.
        e.g. package.subdir.feature_file.MyCallable
        """
        module_path = ".".join(resource_dot_path.split(".")[:-1])
        module = importlib.import_module(module_path)
        return module

    def get_callable_name(self, resource_dot_path: str) -> str:
        """Retrieve the callable based on config string.
        e.g. package.subdir.feature_file.MyCallable
        """
        callable_name = resource_dot_path.split(".")[-1]
        return callable_name

    def get_callable(self, resource_dot_path: str) -> Callable:
        """Retrieve the actual handler class based on config string.
        e.g. package.subdir.feature_file.MyCallable
        """
        module = self.get_module(resource_dot_path)
        callable_name = self.get_callable_name(resource_dot_path)
        return getattr(module, callable_name)


def normalized_pkg_name(pkg_name: str, dashed: bool = False):
    """We maintain consistency by always specifying the package name as
    the "dashed version".

    Python/setuptools will replace "_" with "-" but resource_filename()
    expects the exact directory name, essentially. In order to keep it
    simple upstream and *always* provide package name as the dashed
    version, we do replacement here to 'normalize' both versions to
    whichever convention you need at the time.

    if `dashed`:
        my-package-name --> my-package-name
        my_package_name --> my-package-name

    else:
        my-package-name --> my_package_name
        my_package_name --> my_package_name
    """
    if dashed:
        return str(pkg_name).replace("_", "-")
    return str(pkg_name).replace("-", "_")


def match_prefix(string: str, prefix_p: str) -> bool:
    """For given string, determine whether it begins with provided prefix_p."""
    pattern = re.compile("^(" + prefix_p + ").*")
    if pattern.match(string):
        return True
    return False


def match_suffix(string: str, suffix_p: str) -> bool:
    """For given string, determine whether it ends with provided suffix_p."""
    pattern = re.compile(".*(" + suffix_p + ")$")
    if pattern.match(string):
        return True
    return False


def match_prefix_suffix(string: str, prefix_p: str, suffix_p: str) -> bool:
    """For given string, determine whether it starts w/ prefix & ends w/ suffix"""
    if match_prefix(string, prefix_p) and match_suffix(string, suffix_p):
        return True
    return False


def get_module(resource_dot_path: str):
    """Retrieve the module based on a 'resource dot path'.
    e.g. package.subdir.feature_file.MyCallable
    """
    module_path = ".".join(resource_dot_path.split(".")[:-1])
    module = importlib.import_module(module_path)
    return module


def get_callable_name(resource_dot_path: str) -> str:
    """Retrieve the callable based on config string.
    e.g. package.subdir.feature_file.MyCallable
    """
    callable_name = resource_dot_path.split(".")[-1]
    return callable_name


def get_callable(resource_dot_path: str) -> Callable:
    """Retrieve the actual handler class based on config string.
    e.g. package.subdir.feature_file.MyCallable
    """
    module = get_module(resource_dot_path)
    callable_name = get_callable_name(resource_dot_path)
    return getattr(module, callable_name)



import threading
import signal
import os
import redis
import socket
import sys
import time
import logging
from pypeline.pipelines.middleware.graceful_shutdown_middleware import (
    GraceFulShutdownMiddleware,
)

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def enable_graceful_shutdown(broker, redis_url):
    """Attach GracefulShutdownMiddleware and a SIGTERM handler to the current process."""
    broker.add_middleware(GraceFulShutdownMiddleware(redis_url=redis_url))

    if threading.current_thread().name == "MainThread":
        key_prefix = "busy"
        hostname = socket.gethostname()
        pid = os.getpid()
        busy_key = f"{key_prefix}:{hostname}-{pid}"
        r = redis.Redis.from_url(redis_url)

        def shutdown_handler(signum, frame):
            logger.info(f"[Signal Handler] Received signal {signum}")
            wait_counter = 0
            while r.get(busy_key):
                if wait_counter % 30 == 0:  # Only log every 30 checks
                    logger.info(f"[Signal Handler] Busy ({busy_key}), waiting...")
                time.sleep(1)
                wait_counter += 1
            logger.info(f"[Signal Handler] Done. Exiting.")
            sys.exit(0)

        signal.signal(signal.SIGTERM, shutdown_handler)



import time
import redis
from redis.sentinel import Sentinel
from urllib.parse import urlparse
from pypeline.constants import (
    REDIS_SENTINEL_MASTER_NAME,
    DEFAULT_REDIS_SOCKET_CONNECT_TIMEOUT,
    DEFAULT_REDIS_SOCKET_TIMEOUT,
    DEFAULT_REDIS_RETRY_ON_TIMEOUT,
    DEFAULT_REDIS_SOCKET_KEEPALIVE,
    DEFAULT_REDIS_HEALTH_CHECK_INTERVAL,
)


class LockingParallelBarrier:
    def __init__(self, redis_url, task_key="task_counter", lock_key="task_lock"):
        # Connect to Redis using the provided URL
        if REDIS_SENTINEL_MASTER_NAME is not None:
            parsed_redis_url = urlparse(redis_url)
            redis_sentinel = Sentinel(
                sentinels=[(parsed_redis_url.hostname, parsed_redis_url.port)],
            )
            self.redis = redis_sentinel.master_for(
                REDIS_SENTINEL_MASTER_NAME,
                db=int(parsed_redis_url.path[1]) if parsed_redis_url.path else 0,
                password=parsed_redis_url.password,
                socket_connect_timeout=DEFAULT_REDIS_SOCKET_CONNECT_TIMEOUT,
                socket_timeout=DEFAULT_REDIS_SOCKET_TIMEOUT,
                retry_on_timeout=DEFAULT_REDIS_RETRY_ON_TIMEOUT,
                socket_keepalive=DEFAULT_REDIS_SOCKET_KEEPALIVE,
                health_check_interval=DEFAULT_REDIS_HEALTH_CHECK_INTERVAL,
                decode_responses=True,
            )
        else:
            self.redis = redis.StrictRedis.from_url(redis_url, decode_responses=True)
        self.task_key = task_key
        self.lock_key = lock_key

    def acquire_lock(self, timeout=5):
        """Acquire a lock using Redis."""
        while True:
            if self.redis.set(self.lock_key, "locked", nx=True, ex=timeout):
                return True
            time.sleep(0.1)

    def release_lock(self):
        """Release the lock in Redis."""
        self.redis.delete(self.lock_key)

    def set_task_count(self, count):
        """Initialize the task counter in Redis."""
        self.redis.set(self.task_key, count)

    def decrement_task_count(self):
        """Decrement the task counter in Redis."""
        return self.redis.decr(self.task_key)

    def task_exists(self):
        return self.redis.exists(self.task_key)

    def get_task_count(self):
        """Get the current value of the task counter."""
        return int(self.redis.get(self.task_key) or 0)
